@@ -224,9 +224,6 @@ gesture_unregister_point (ClutterGestureAction *action, gint position)
 {
   ClutterGestureActionPrivate *priv = action->priv;
 
-  if (action->priv->points->len == 0)
-    return;
-
   g_array_remove_index (priv->points, position);
 }
 
@@ -355,8 +352,8 @@ begin_gesture (ClutterGestureAction *action,
 }
 
 static gboolean
-stage_captured_event_cb (ClutterActor         *stage,
-                         ClutterEvent         *event,
+stage_captured_event_cb (ClutterActor       *stage,
+                         ClutterEvent       *event,
                          ClutterGestureAction *action)
 {
   ClutterGestureActionPrivate *priv = action->priv;
@@ -365,20 +362,6 @@ stage_captured_event_cb (ClutterActor         *stage,
   float threshold_x, threshold_y;
   gboolean return_value;
   GesturePoint *point;
-  ClutterEventType event_type;
-
-  event_type = clutter_event_type (event);
-  if (event_type != CLUTTER_TOUCH_CANCEL &&
-      event_type != CLUTTER_TOUCH_UPDATE &&
-      event_type != CLUTTER_TOUCH_END &&
-      event_type != CLUTTER_MOTION &&
-      event_type != CLUTTER_BUTTON_RELEASE)
-    return CLUTTER_EVENT_PROPAGATE;
-
-  if (event->type != CLUTTER_TOUCH_CANCEL &&
-      event->type != CLUTTER_TOUCH_UPDATE && event->type != CLUTTER_TOUCH_END &&
-      event->type != CLUTTER_MOTION && event->type != CLUTTER_BUTTON_RELEASE)
-    return CLUTTER_EVENT_PROPAGATE;
 
   if ((point = gesture_find_point (action, event, &position)) == NULL)
     return CLUTTER_EVENT_PROPAGATE;
@@ -488,7 +471,7 @@ stage_captured_event_cb (ClutterActor         *stage,
       break;
     }
 
-  if (priv->points->len == 0 && priv->stage_capture_id)
+  if (priv->points->len == 0)
     {
       g_signal_handler_disconnect (priv->stage, priv->stage_capture_id);
       priv->stage_capture_id = 0;
@@ -1267,7 +1250,7 @@ clutter_gesture_action_set_threshold_trigger_edge (ClutterGestureAction      *ac
 }
 
 /**
- * clutter_gesture_action_get_threshold_trigger_egde:
+ * clutter_gesture_action_get_threshold_trigger_edge:
  * @action: a #ClutterGestureAction
  *
  * Retrieves the edge trigger of the gesture @action, as set using
@@ -1278,7 +1261,7 @@ clutter_gesture_action_set_threshold_trigger_edge (ClutterGestureAction      *ac
  * Since: 1.18
  */
 ClutterGestureTriggerEdge
-clutter_gesture_action_get_threshold_trigger_egde (ClutterGestureAction *action)
+clutter_gesture_action_get_threshold_trigger_edge (ClutterGestureAction *action)
 {
   g_return_val_if_fail (CLUTTER_IS_GESTURE_ACTION (action),
                         CLUTTER_GESTURE_TRIGGER_EDGE_NONE);
