@@ -62,6 +62,15 @@
 #define INITIAL_POINTER_X 16
 #define INITIAL_POINTER_Y 16
 
+typedef struct _ClutterTouchState ClutterTouchState;
+typedef struct _ClutterEventFilter ClutterEventFilter;
+
+struct _ClutterTouchState
+{
+  guint32 id;
+  ClutterPoint coords;
+};
+
 struct _ClutterSeatEvdev
 {
   struct libinput_seat *libinput_seat;
@@ -72,6 +81,8 @@ struct _ClutterSeatEvdev
   ClutterInputDevice *core_pointer;
   ClutterInputDevice *core_keyboard;
 
+  GHashTable *touches;
+
   struct xkb_state *xkb;
   xkb_led_index_t caps_lock_led;
   xkb_led_index_t num_lock_led;
@@ -88,6 +99,13 @@ struct _ClutterSeatEvdev
   ClutterInputDevice *repeat_device;
 };
 
+struct _ClutterEventFilter
+{
+  ClutterEvdevFilterFunc func;
+  gpointer data;
+  GDestroyNotify destroy_notify;
+};
+
 typedef struct _ClutterEventSource  ClutterEventSource;
 
 struct _ClutterDeviceManagerEvdevPrivate
@@ -111,6 +129,8 @@ struct _ClutterDeviceManagerEvdevPrivate
   ClutterStageManager *stage_manager;
   guint stage_added_handler;
   guint stage_removed_handler;
+
+  GSList *event_filters;
 };
 
 G_DEFINE_TYPE_WITH_PRIVATE (ClutterDeviceManagerEvdev,
@@ -121,6 +141,7 @@ static ClutterOpenDeviceCallback  device_open_callback;
 static ClutterCloseDeviceCallback device_close_callback;
 static gpointer                   device_callback_data;
 
+#ifdef CLUTTER_ENABLE_DEBUG
 static const char *device_type_str[] = {
   "pointer",            /* CLUTTER_POINTER_DEVICE */
   "keyboard",           /* CLUTTER_KEYBOARD_DEVICE */
@@ -133,6 +154,7 @@ static const char *device_type_str[] = {
   "eraser",             /* CLUTTER_ERASER_DEVICE */
   "cursor",             /* CLUTTER_CURSOR_DEVICE */
 };
+#endif /* CLUTTER_ENABLE_DEBUG */
 
 /*
  * ClutterEventSource management
@@ -367,8 +389,11 @@ notify_absolute_motion (ClutterInputDevice *input_device,
   _clutter_xkb_translate_state (event, seat->xkb, seat->button_state);
   event->motion.x = x;
   event->motion.y = y;
+  clutter_event_set_device (event, seat->core_pointer);
   clutter_event_set_source_device (event, input_device);
 
+  _clutter_input_device_set_stage (seat->core_pointer, stage);
+
   queue_event (event);
 }
 
@@ -417,7 +442,7 @@ notify_scroll (ClutterInputDevice *input_device,
   ClutterStage *stage;
   ClutterEvent *event = NULL;
   ClutterPoint point;
-  const gdouble scroll_factor = 10.0f;
+  gdouble scroll_factor;
 
   /* We can drop the event on the floor if no stage has been
    * associated with the device yet. */
@@ -435,7 +460,11 @@ notify_scroll (ClutterInputDevice *input_device,
   event->scroll.device = seat->core_pointer;
   _clutter_xkb_translate_state (event, seat->xkb, seat->button_state);
 
+  /* libinput pointer axis events are in pointer motion coordinate space.
+   * To convert to Xi2 discrete step coordinate space, multiply the factor
+   * 1/10. */
   event->scroll.direction = CLUTTER_SCROLL_SMOOTH;
+  scroll_factor = 1.0 / 10.0;
   clutter_event_set_scroll_delta (event,
                                   scroll_factor * dx,
                                   scroll_factor * dy);
@@ -443,6 +472,7 @@ notify_scroll (ClutterInputDevice *input_device,
   clutter_input_device_get_coords (seat->core_pointer, NULL, &point);
   event->scroll.x = point.x;
   event->scroll.y = point.y;
+  clutter_event_set_device (event, seat->core_pointer);
   clutter_event_set_source_device (event, input_device);
 
   queue_event (event);
@@ -522,6 +552,52 @@ notify_button (ClutterInputDevice *input_device,
   clutter_input_device_get_coords (seat->core_pointer, NULL, &point);
   event->button.x = point.x;
   event->button.y = point.y;
+  clutter_event_set_device (event, seat->core_pointer);
+  clutter_event_set_source_device (event, input_device);
+
+  _clutter_input_device_set_stage (seat->core_pointer, stage);
+
+  queue_event (event);
+}
+
+static void
+notify_touch_event (ClutterInputDevice *input_device,
+		    ClutterEventType    evtype,
+		    guint32             time_,
+		    gint32              slot,
+		    gdouble             x,
+		    gdouble             y)
+{
+  ClutterInputDeviceEvdev *device_evdev;
+  ClutterSeatEvdev *seat;
+  ClutterStage *stage;
+  ClutterEvent *event = NULL;
+
+  /* We can drop the event on the floor if no stage has been
+   * associated with the device yet. */
+  stage = _clutter_input_device_get_stage (input_device);
+  if (!stage)
+    return;
+
+  device_evdev = CLUTTER_INPUT_DEVICE_EVDEV (input_device);
+  seat = _clutter_input_device_evdev_get_seat (device_evdev);
+
+  event = clutter_event_new (evtype);
+
+  event->touch.time = time_;
+  event->touch.stage = CLUTTER_STAGE (stage);
+  event->touch.device = seat->core_pointer;
+  event->touch.x = x;
+  event->touch.y = y;
+  /* "NULL" sequences are special cased in clutter */
+  event->touch.sequence = GINT_TO_POINTER (slot + 1);
+  _clutter_xkb_translate_state (event, seat->xkb, seat->button_state);
+
+  if (evtype == CLUTTER_TOUCH_BEGIN ||
+      evtype == CLUTTER_TOUCH_UPDATE)
+    event->touch.modifier_state |= CLUTTER_BUTTON1_MASK;
+
+  clutter_event_set_device (event, seat->core_pointer);
   clutter_event_set_source_device (event, input_device);
 
   queue_event (event);
@@ -577,8 +653,7 @@ clutter_event_dispatch (GSource     *g_source,
       _clutter_stage_queue_event (event->any.stage, event, FALSE);
 
       /* update the device states *after* the event */
-      event_state = seat->button_state |
-        xkb_state_serialize_mods (seat->xkb, XKB_STATE_MODS_EFFECTIVE);
+      event_state = xkb_state_serialize_mods (seat->xkb, XKB_STATE_MODS_EFFECTIVE);
       _clutter_input_device_set_state (seat->core_pointer, event_state);
       _clutter_input_device_set_state (seat->core_keyboard, event_state);
     }
@@ -637,6 +712,12 @@ clutter_event_source_free (ClutterEventSource *source)
   g_source_unref (g_source);
 }
 
+static void
+clutter_touch_state_free (ClutterTouchState *touch_state)
+{
+  g_slice_free (ClutterTouchState, touch_state);
+}
+
 static void
 clutter_seat_evdev_set_libinput_seat (ClutterSeatEvdev *seat,
                                       struct libinput_seat *libinput_seat)
@@ -676,6 +757,9 @@ clutter_seat_evdev_new (ClutterDeviceManagerEvdev *manager_evdev)
   _clutter_device_manager_add_device (manager, device);
   seat->core_keyboard = device;
 
+  seat->touches = g_hash_table_new_full (NULL, NULL, NULL,
+                                         (GDestroyNotify) clutter_touch_state_free);
+
   ctx = xkb_context_new(0);
   g_assert (ctx);
 
@@ -721,6 +805,7 @@ clutter_seat_evdev_free (ClutterSeatEvdev *seat)
       g_object_unref (device);
     }
   g_slist_free (seat->devices);
+  g_hash_table_unref (seat->touches);
 
   xkb_state_unref (seat->xkb);
 
@@ -969,6 +1054,45 @@ process_base_event (ClutterDeviceManagerEvdev *manager_evdev,
   return handled;
 }
 
+static ClutterTouchState *
+_device_seat_add_touch (ClutterInputDevice *input_device,
+                        guint32             id)
+{
+  ClutterInputDeviceEvdev *device_evdev =
+    CLUTTER_INPUT_DEVICE_EVDEV (input_device);
+  ClutterSeatEvdev *seat = _clutter_input_device_evdev_get_seat (device_evdev);
+  ClutterTouchState *touch;
+
+  touch = g_slice_new0 (ClutterTouchState);
+  touch->id = id;
+
+  g_hash_table_insert (seat->touches, GUINT_TO_POINTER (id), touch);
+
+  return touch;
+}
+
+static void
+_device_seat_remove_touch (ClutterInputDevice *input_device,
+                           guint32             id)
+{
+  ClutterInputDeviceEvdev *device_evdev =
+    CLUTTER_INPUT_DEVICE_EVDEV (input_device);
+  ClutterSeatEvdev *seat = _clutter_input_device_evdev_get_seat (device_evdev);
+
+  g_hash_table_remove (seat->touches, GUINT_TO_POINTER (id));
+}
+
+static ClutterTouchState *
+_device_seat_get_touch (ClutterInputDevice *input_device,
+                        guint32             id)
+{
+  ClutterInputDeviceEvdev *device_evdev =
+    CLUTTER_INPUT_DEVICE_EVDEV (input_device);
+  ClutterSeatEvdev *seat = _clutter_input_device_evdev_get_seat (device_evdev);
+
+  return g_hash_table_lookup (seat->touches, GUINT_TO_POINTER (id));
+}
+
 static gboolean
 process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
                       struct libinput_event *event)
@@ -1090,6 +1214,119 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
 
       }
 
+    case LIBINPUT_EVENT_TOUCH_DOWN:
+      {
+        gint32 slot;
+        guint32 time;
+        li_fixed_t x, y;
+        gfloat stage_width, stage_height;
+        ClutterStage *stage;
+        ClutterTouchState *touch_state;
+        struct libinput_event_touch *touch_event =
+          libinput_event_get_touch_event (event);
+        device = libinput_device_get_user_data (libinput_device);
+
+        stage = _clutter_input_device_get_stage (device);
+        if (!stage)
+          break;
+
+        stage_width = clutter_actor_get_width (CLUTTER_ACTOR (stage));
+        stage_height = clutter_actor_get_height (CLUTTER_ACTOR (stage));
+
+        slot = libinput_event_touch_get_slot (touch_event);
+        time = libinput_event_touch_get_time (touch_event);
+        x = libinput_event_touch_get_x_transformed (touch_event,
+                                                    stage_width);
+        y = libinput_event_touch_get_y_transformed (touch_event,
+                                                    stage_height);
+
+        touch_state = _device_seat_add_touch (device, slot);
+        touch_state->coords.x = li_fixed_to_double (x);
+        touch_state->coords.y = li_fixed_to_double (y);
+
+        notify_touch_event (device, CLUTTER_TOUCH_BEGIN, time, slot,
+                             touch_state->coords.x, touch_state->coords.y);
+        break;
+      }
+
+    case LIBINPUT_EVENT_TOUCH_UP:
+      {
+        gint32 slot;
+        guint32 time;
+        ClutterTouchState *touch_state;
+        struct libinput_event_touch *touch_event =
+          libinput_event_get_touch_event (event);
+        device = libinput_device_get_user_data (libinput_device);
+
+        slot = libinput_event_touch_get_slot (touch_event);
+        time = libinput_event_touch_get_time (touch_event);
+        touch_state = _device_seat_get_touch (device, slot);
+
+        notify_touch_event (device, CLUTTER_TOUCH_END, time, slot,
+			    touch_state->coords.x, touch_state->coords.y);
+        _device_seat_remove_touch (device, slot);
+
+        break;
+      }
+
+    case LIBINPUT_EVENT_TOUCH_MOTION:
+      {
+        gint32 slot;
+        guint32 time;
+        li_fixed_t x, y;
+        gfloat stage_width, stage_height;
+        ClutterStage *stage;
+        ClutterTouchState *touch_state;
+        struct libinput_event_touch *touch_event =
+          libinput_event_get_touch_event (event);
+        device = libinput_device_get_user_data (libinput_device);
+
+        stage = _clutter_input_device_get_stage (device);
+        if (!stage)
+          break;
+
+        stage_width = clutter_actor_get_width (CLUTTER_ACTOR (stage));
+        stage_height = clutter_actor_get_height (CLUTTER_ACTOR (stage));
+
+        slot = libinput_event_touch_get_slot (touch_event);
+        time = libinput_event_touch_get_time (touch_event);
+        x = libinput_event_touch_get_x_transformed (touch_event,
+                                                    stage_width);
+        y = libinput_event_touch_get_y_transformed (touch_event,
+                                                    stage_height);
+
+        touch_state = _device_seat_get_touch (device, slot);
+        touch_state->coords.x = li_fixed_to_double (x);
+        touch_state->coords.y = li_fixed_to_double (y);
+
+        notify_touch_event (device, CLUTTER_TOUCH_UPDATE, time, slot,
+			    touch_state->coords.x, touch_state->coords.y);
+        break;
+      }
+    case LIBINPUT_EVENT_TOUCH_CANCEL:
+      {
+        ClutterTouchState *touch_state;
+        GHashTableIter iter;
+        guint32 time;
+        struct libinput_event_touch *touch_event =
+          libinput_event_get_touch_event (event);
+        ClutterSeatEvdev *seat;
+
+        device = libinput_device_get_user_data (libinput_device);
+        time = libinput_event_touch_get_time (touch_event);
+        seat = _clutter_input_device_evdev_get_seat (CLUTTER_INPUT_DEVICE_EVDEV (device));
+        g_hash_table_iter_init (&iter, seat->touches);
+
+        while (g_hash_table_iter_next (&iter, NULL, (gpointer*) &touch_state))
+          {
+            notify_touch_event (device, CLUTTER_TOUCH_CANCEL,
+                                time, touch_state->id,
+                                touch_state->coords.x, touch_state->coords.y);
+            g_hash_table_iter_remove (&iter);
+          }
+
+        break;
+      }
     default:
       handled = FALSE;
     }
@@ -1097,10 +1334,40 @@ process_device_event (ClutterDeviceManagerEvdev *manager_evdev,
   return handled;
 }
 
+static gboolean
+filter_event (ClutterDeviceManagerEvdev *manager_evdev,
+              struct libinput_event     *event)
+{
+  gboolean retval = CLUTTER_EVENT_PROPAGATE;
+  ClutterEventFilter *filter;
+  GSList *tmp_list;
+
+  tmp_list = manager_evdev->priv->event_filters;
+
+  while (tmp_list)
+    {
+      filter = tmp_list->data;
+      retval = filter->func (event, filter->data);
+      tmp_list = tmp_list->next;
+
+      if (retval != CLUTTER_EVENT_PROPAGATE)
+        break;
+    }
+
+  return retval;
+}
+
 static void
 process_event (ClutterDeviceManagerEvdev *manager_evdev,
                struct libinput_event *event)
 {
+  gboolean retval;
+
+  retval = filter_event (manager_evdev, event);
+
+  if (retval != CLUTTER_EVENT_PROPAGATE)
+    return;
+
   if (process_base_event (manager_evdev, event))
     return;
   if (process_device_event (manager_evdev, event))
@@ -1638,3 +1905,94 @@ clutter_evdev_set_keyboard_repeat (ClutterDeviceManager *evdev,
   seat->repeat_delay = delay;
   seat->repeat_interval = interval;
 }
+
+/**
+ * clutter_evdev_add_filter: (skip)
+ * @func: (closure data): a filter function
+ * @data: (allow-none): user data to be passed to the filter function, or %NULL
+ * @destroy_notify: (allow-none): function to call on @data when the filter is removed, or %NULL
+ *
+ * Adds an event filter function.
+ *
+ * Since: 1.20
+ * Stability: unstable
+ */
+void
+clutter_evdev_add_filter (ClutterEvdevFilterFunc func,
+                          gpointer               data,
+                          GDestroyNotify         destroy_notify)
+{
+  ClutterDeviceManagerEvdev *manager_evdev;
+  ClutterDeviceManager *manager;
+  ClutterEventFilter *filter;
+
+  g_return_if_fail (func != NULL);
+
+  manager = clutter_device_manager_get_default ();
+
+  if (!CLUTTER_IS_DEVICE_MANAGER_EVDEV (manager))
+    {
+      g_critical ("The Clutter input backend is not a evdev backend");
+      return;
+    }
+
+  manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (manager);
+
+  filter = g_new0 (ClutterEventFilter, 1);
+  filter->func = func;
+  filter->data = data;
+  filter->destroy_notify = destroy_notify;
+
+  manager_evdev->priv->event_filters =
+    g_slist_append (manager_evdev->priv->event_filters, filter);
+}
+
+/**
+ * clutter_evdev_remove_filter: (skip)
+ * @func: a filter function
+ * @data: (allow-none): user data to be passed to the filter function, or %NULL
+ *
+ * Removes the given filter function.
+ *
+ * Since: 1.20
+ * Stability: unstable
+ */
+void
+clutter_evdev_remove_filter (ClutterEvdevFilterFunc func,
+                             gpointer               data)
+{
+  ClutterDeviceManagerEvdev *manager_evdev;
+  ClutterDeviceManager *manager;
+  ClutterEventFilter *filter;
+  GSList *tmp_list;
+
+  g_return_if_fail (func != NULL);
+
+  manager = clutter_device_manager_get_default ();
+
+  if (!CLUTTER_IS_DEVICE_MANAGER_EVDEV (manager))
+    {
+      g_critical ("The Clutter input backend is not a evdev backend");
+      return;
+    }
+
+  manager_evdev = CLUTTER_DEVICE_MANAGER_EVDEV (manager);
+  tmp_list = manager_evdev->priv->event_filters;
+
+  while (tmp_list)
+    {
+      filter = tmp_list->data;
+
+      if (filter->func == func && filter->data == data)
+        {
+          if (filter->destroy_notify)
+            filter->destroy_notify (filter->data);
+          g_free (filter);
+          manager_evdev->priv->event_filters =
+            g_slist_delete_link (manager_evdev->priv->event_filters, tmp_list);
+          return;
+        }
+
+      tmp_list = tmp_list->next;
+    }
+}
