@@ -28,6 +28,7 @@
 #include <glib-object.h>
 #include <xkbcommon/xkbcommon.h>
 #include <clutter/clutter.h>
+#include <libinput.h>
 
 G_BEGIN_DECLS
 
@@ -102,6 +103,22 @@ void clutter_evdev_set_keyboard_repeat (ClutterDeviceManager *evdev,
                                         guint32               delay,
                                         guint32               interval);
 
+typedef gboolean (* ClutterEvdevFilterFunc) (struct libinput_event *event,
+                                             gpointer               data);
+
+CLUTTER_AVAILABLE_IN_1_20
+void clutter_evdev_add_filter    (ClutterEvdevFilterFunc func,
+                                  gpointer               data,
+                                  GDestroyNotify         destroy_notify);
+CLUTTER_AVAILABLE_IN_1_20
+void clutter_evdev_remove_filter (ClutterEvdevFilterFunc func,
+                                  gpointer               data);
+CLUTTER_AVAILABLE_IN_1_20
+struct libinput_device * clutter_evdev_input_device_get_libinput_device (ClutterInputDevice *device);
+
+CLUTTER_AVAILABLE_IN_1_20
+gint32 clutter_evdev_event_sequence_get_slot (const ClutterEventSequence *sequence);
+
 G_END_DECLS
 
 #endif /* __CLUTTER_EVDEV_H__ */
